import collections.abc
import pandas as pd
import pathlib
import typing
import sipametrics.models.direct_alpha as direct_alpha_models


def _load_csv(file_path: typing.Union[str, pathlib.Path], field_names: typing.Optional[list[str]]=None, date_fields: typing.Optional[list[str]]=None) -> pd.DataFrame:
    kwargs = {}
    if field_names:
        kwargs["usecols"] = field_names
    if date_fields:
        kwargs["parse_dates"] = date_fields

    df = pd.read_csv(file_path, **kwargs)
    return df

def _sniff_csv(file_path: typing.Union[str, pathlib.Path], field_names: list[str]) -> list[str]: 
    with open(file_path, "r") as f:
        headline = f.readline().strip().split(",")

    headers = { column.lower() for column in headline }
    columns = [ column for column in field_names if column.lower() in headers ]
    return columns
    

def load_fund_datapoints(file_path: typing.Union[str, pathlib.Path]) -> list[direct_alpha_models.FundDatapoint]:
    """
    Loads fund datapoints from a CSV file, and returns the rows as a list of FundDatapoint objects. Note that first row of the file must contain the following fields:
    date, contribution, distribution.

    Args:
        file_path (str, Path): Path to the CSV file.

    Returns:
        list (FundDatapoint): List of FundDatapoint objects, encapsulating entries of date, contribution and distribution.
    """
    field_names = _sniff_csv(file_path, field_names=["date", "contribution", "distribution"])
    date_fields = ["date"]

    df = _load_csv(file_path=file_path, field_names=field_names, date_fields=date_fields)
    records = typing.cast(collections.abc.Sequence[dict[str, typing.Any]], df.to_dict(orient="records"))
    datapoints = [direct_alpha_models.FundDatapoint(**row) for row in records]
    return datapoints

def load_index_benchmark_datapoints(file_path: typing.Union[str, pathlib.Path]) -> list[direct_alpha_models.IndexBenchmarkDatapoint]:
    """
    Loads index benchmark datapoints from a CSV file, and returns the rows as a list of IndexBenchmarkDatapoint objects. Note that first row of the file must contain
    the following fields: date, level.

    Args:
        file_path (str, Path): Path to the CSV file.

    Returns:
        list (IndexBenchmarkDatapoint): List of IndexBenchmarkDatapoint objects, encapsulating entries of date and (index) level.
    """
    field_names = _sniff_csv(file_path, field_names=["date", "level"])
    date_fields = ["date"]

    df = _load_csv(file_path=file_path, field_names=field_names, date_fields=date_fields)
    records = typing.cast(collections.abc.Sequence[dict[str, typing.Any]], df.to_dict(orient="records"))
    datapoints = [direct_alpha_models.IndexBenchmarkDatapoint(**row) for row in records]
    return datapoints



import asyncio
import aiohttp
import logging
import sipametrics.constants as CONSTANTS
import sipametrics.internal.endpoints as endpoints


class BaseClient:
    def __init__(self, api_key: str, api_secret: str):
        self._session = aiohttp.ClientSession()
        self._api_key = api_key
        self._api_secret = api_secret

    async def close(self):
        await self._session.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, tb):
        await self._session.close()

    async def _get(self, path, **kwargs):
        return await self._request("GET", path, **kwargs)

    async def _post(self, path, **kwargs) -> dict:
        return await self._request("POST", path, **kwargs)

    def _generate_authorization_headers(self, method) -> dict:
        if method == "GET":
            return {
                "X-API-KEY": self._api_key,
                "X-API-SECRET": self._api_secret,
                "X-Source": f"{endpoints.SOURCE}",
                "X-Version": f"{endpoints.VERSION}"
            }
        elif method == "POST":
            return {
                "X-API-KEY": self._api_key,
                "X-API-SECRET": self._api_secret,
                "X-Source": f"{endpoints.SOURCE}",
                "X-Version": f"{endpoints.VERSION}",
                "Content-Type": "application/json",
            }
        else:
            return {}

    async def _request(self, method: str, uri: str, **kwargs) -> dict:
        headers = {}

        if method in ["POST", "GET"]:
            headers.update(self._generate_authorization_headers(method))

        logging.debug(f"{CONSTANTS.IDENTIFIER}: {method}; {uri}; {kwargs}")

        query_params = kwargs.pop("params", None)

        async with getattr(self._session, method.lower())(
            uri,
            params=query_params,
            headers=headers,
            json=kwargs.get("data"),
        ) as response:
            return await self._handle_response(response)

    async def _handle_response(self, response: aiohttp.ClientResponse) -> dict:
        """
        Handles retries for transient errors and processes API responses.
        """
        retries = 3
        delay = 2
        for _ in range(retries):
            if response.status == 200:
                try:
                    return await response.json()
                except aiohttp.ContentTypeError as e:
                    raise ValueError("Invalid JSON response") from e
            elif response.status in [500, 502, 503, 504]:
                await asyncio.sleep(delay)
                delay *= 2
            else:
                response.raise_for_status()

        raise Exception(f"Failed to process response after {retries} retries. Status: {response.status}")



import datetime
import pydantic
import typing
import sipametrics.models.base as base_models
import sipametrics.utilities as utilities
import sipametrics.constants as CONSTANTS


class TiccsProfile(base_models.CamelCaseBaseModel):
    industrial_activities: typing.Optional[list[str]] = None
    business_risk: typing.Optional[str] = pydantic.Field(default=None, serialization_alias="businessModel")
    corporate_structure: typing.Optional[str] = None

    def to_list(self) -> list:
        entries = []
        for key,value in self.model_dump(by_alias=True, exclude_defaults=True).items():
            entries.append({ "profile": { key: value } })

        return entries
    
class FactorProfile(base_models.CamelCaseBaseModel):
    currency: typing.Optional[str] = None
    countries: typing.Optional[list[str]] = None

    # country_ids: typing.Optional[list[str]] = pydantic.Field(default=None, exclude=False, validate_default=False)

    @pydantic.field_validator("currency", mode="after")
    @classmethod
    def validate_currency(cls, v: typing.Optional[str]) -> typing.Optional[str]:
        if v and v.upper() not in CONSTANTS.CURRENCY_MAP:
            raise ValueError(f"Invalid currency code: {v}")

        return v
    
    @pydantic.field_validator("countries", mode="after")
    @classmethod
    def validate_countries(cls, v: typing.Optional[list[str]]) -> typing.Optional[list[str]]:
        if v:
            for country in v:
                if country not in CONSTANTS.COUNTRY_MAP:
                    raise ValueError(f"Invalid country code: {country}")
            
            # cls.country_ids = [ CONSTANTS.COUNTRY_MAP[country] for country in values.countries ]

        return v

class EquityFactorProfile(FactorProfile):
    size: typing.Optional[typing.Union[float, str]] = None
    leverage: typing.Optional[typing.Union[float, str]] = None
    profitability: typing.Optional[typing.Union[float, str]] = None
    investment: typing.Optional[typing.Union[float, str]] = None
    time_to_maturity: typing.Optional[typing.Union[float, str]] = None

    @pydantic.field_validator("time_to_maturity", mode="after")
    @classmethod
    def validate_time_to_maturity(cls, v: typing.Optional[typing.Union[float, str]]) -> typing.Optional[typing.Union[float, str]]:
        if isinstance(v, str) and v not in CONSTANTS.TIME_TO_MATURITY_BUCKETS:
            raise ValueError(f"Invalid time to maturity bucket: {v}")

        return v

    def to_list(self) -> list:
        entries = []

        if self.countries:
            profile = { 
                "profile": {
                    "factor": "countries",
                    "countries": [ CONSTANTS.COUNTRY_MAP[country] for country in self.countries ]
                }
            }
            entries.append(profile)

        for key,value in self.model_dump(by_alias=True, exclude_defaults=True, exclude={"currency", "countries"}).items():
            profile = { "factor": key, key: value }

            if isinstance(value, str) and len(value) > 1:
                if key == "timeToMaturity":
                    profile["timeToMaturityBucket"] = CONSTANTS.TIME_TO_MATURITY_BUCKETS[value]
                else:
                    profile["quintile"] = int(value[1:])
                profile.pop(key)

            if key in ["size"] and self.currency:
                profile["currency"] = CONSTANTS.CURRENCY_MAP[self.currency.upper()]

            entries.append({ "profile": profile })

        return entries
    
class DebtFactorProfile(FactorProfile):
    face_value: typing.Optional[typing.Union[float, str]] = None
    time_to_maturity: typing.Optional[typing.Union[float, str]] = pydantic.Field(default=None, serialization_alias="debtTimeToMaturity")

    @pydantic.field_validator("time_to_maturity", mode="after")
    @classmethod
    def validate_time_to_maturity(cls, v: typing.Optional[typing.Union[float, str]]) -> typing.Optional[typing.Union[float, str]]:
        if isinstance(v, str) and v not in CONSTANTS.TIME_TO_MATURITY_BUCKETS:
            raise ValueError(f"Invalid time to maturity bucket: {v}")

        return v

    def to_list(self) -> list:
        entries = []

        if self.countries:
            profile = { 
                "profile": {
                    "factor": "countries",
                    "countries": [ CONSTANTS.COUNTRY_MAP[country] for country in self.countries ]
                }
            }
            entries.append(profile)

        for key,value in self.model_dump(by_alias=True, exclude_defaults=True, exclude={"currency", "countries"}).items():
            profile = { "factor": key, key: value }

            if isinstance(value, str) and len(value) > 1:
                if key == "debtTimeToMaturity":
                    profile["timeToMaturityBucket"] = CONSTANTS.TIME_TO_MATURITY_BUCKETS[value]
                else:
                    profile["quintile"] = int(value[1:])
                profile.pop(key)

            if key in ["faceValue"] and self.currency:
                profile["currency"] = CONSTANTS.CURRENCY_MAP[self.currency.upper()]

            entries.append({ "profile": profile })

        return entries    

class ComparableRequest(base_models.CamelCaseBaseModel):
    metric: str
    age_in_months: typing.Optional[int] = pydantic.Field(default=None, serialization_alias="age")
    end_date: typing.Optional[datetime.date] = None
    window_in_years: typing.Optional[int] = pydantic.Field(default=None, exclude=True, gt=0)
    ticcs_profiles: typing.Optional[TiccsProfile] = None 
    factors_profiles: typing.Optional[typing.Union[EquityFactorProfile, DebtFactorProfile]] = None
    intersect_ticcs: typing.Optional[bool] = None   
    factor_weight: typing.Optional[float] = None

    start_date: typing.Optional[datetime.date] = pydantic.Field(default=None, exclude=False, validate_default=False)

    @pydantic.model_validator(mode="after")
    def compute_start_date(cls, values):
        if values.age_in_months is not None:
            values.start_date = None
            values.end_date = None

            return values
        
        if values.end_date is None or values.window_in_years is None:
            raise ValueError(
                "Either 'age_in_months' must be set, "
                "or both 'end_date' and 'window_in_years' must be provided."
            )

        values.start_date = utilities.calculate_start_date(end_date=values.end_date, window_in_years=values.window_in_years)
        return values

    def to_dict(self) -> dict:
        payload = self.model_dump(mode="json", by_alias=True, exclude={"ticcs_profiles", "factors_profiles"}, exclude_defaults=True)
        if self.ticcs_profiles:
            payload["ticcsProfiles"] = self.ticcs_profiles.to_list()
        if self.factors_profiles:
            payload["factorsProfiles"] = self.factors_profiles.to_list()

        return payload


import datetime
import pydantic
import sipametrics.models.base as base_models
import sipametrics.constants as CONSTANTS


class TermStructureRequest(base_models.CamelCaseBaseModel):
    country: str = pydantic.Field(serialization_alias="countryName")
    date: datetime.date = pydantic.Field(serialization_alias="valueDate")
    maturity_date: datetime.date

    @pydantic.field_validator("country", mode="after")
    @classmethod
    def validate_countries(cls, v: str) -> str:
        if v not in CONSTANTS.COUNTRY_MAP:
            raise ValueError(f"Invalid country code: {v}")

        return v

    def to_dict(self) -> dict:
        return self.model_dump(mode="json", by_alias=True)


import datetime
import pydantic
import typing
import sipametrics.models.base as base_models
import sipametrics.utilities as utilities
import sipametrics.constants as CONSTANTS


class PeccsCountryProfile(base_models.CamelCaseBaseModel):
    industrial_activities: typing.Optional[list[str]] = None
    revenue_models: typing.Optional[list[str]] = None
    customer_models: typing.Optional[list[str]] = None
    lifecycle_phases: typing.Optional[list[str]] = None
    value_chain_types: typing.Optional[list[str]] = None
    countries: typing.Optional[list[str]] = None

    @pydantic.field_validator("countries", mode="after")
    @classmethod
    def validate_countries(cls, v: typing.Optional[list[str]]) -> typing.Optional[list[str]]:
        if v:
            for country in v:
                if country not in CONSTANTS.COUNTRY_MAP:
                    raise ValueError(f"Invalid country code: {country}")

        return v
       
class FactorProfile(base_models.CamelCaseBaseModel):
    size: typing.Optional[typing.Union[float, str]] = None
    growth: typing.Optional[typing.Union[float, str]] = None
    leverage: typing.Optional[typing.Union[float, str]] = None
    profits: typing.Optional[typing.Union[float, str]] = None
    country_risk: typing.Optional[list[str]] = pydantic.Field(default=None, serialization_alias="termSpread")

    @pydantic.field_validator("country_risk", mode="after")
    @classmethod
    def validate_countries(cls, v: typing.Optional[list[str]]) -> typing.Optional[list[str]]:
        if v:
            for country in v:
                if country not in CONSTANTS.COUNTRY_MAP:
                    raise ValueError(f"Invalid country code: {country}")

        return v

    def to_list(self, currency: typing.Optional[str] = None) -> list:
        entries = []

        for key,value in self.model_dump(by_alias=True, exclude_defaults=True).items():
            profile = { "factorName": key.title(), key: value }

            if (key == "termSpread"):
                profile["countries"] = value
                profile.pop(key)

                entries.append(profile)
                continue

            if isinstance(value, str) and len(value) > 1:
                profile["quintile"] = int(value[1:])
                profile.pop(key)

            if (key == "size" and currency):
                profile["currency"] = currency

            entries.append(profile)

        return entries    
    
class ComparableBaseRequest(base_models.CamelCaseBaseModel):
    metric: str
    currency: typing.Optional[str] = pydantic.Field(default=None, serialization_alias="reportingCurrency")
    age_in_months: typing.Optional[int] = pydantic.Field(default=None, serialization_alias="age")
    end_date: typing.Optional[datetime.date] = None
    window_in_years: typing.Optional[int] = pydantic.Field(default=None, exclude=True, gt=0)
    universe: typing.Optional[str] = None
    factor_weight: typing.Optional[float] = None

    start_date: typing.Optional[datetime.date] = pydantic.Field(default=None, exclude=False, validate_default=False)

    @pydantic.field_validator("currency", mode="after")
    @classmethod
    def validate_currency(cls, v: typing.Optional[str]) -> typing.Optional[str]:
        if v and v.upper() not in CONSTANTS.CURRENCY_MAP:
            raise ValueError(f"Invalid currency code: {v}")

        return v

    @pydantic.model_validator(mode="after")
    def compute_start_date(cls, values):
        if values.age_in_months:
            values.start_date = None
            values.end_date = None

            return values
        
        if not (values.end_date and values.window_in_years):
            raise ValueError(
                "Either 'age_in_months' must be set, "
                "or both 'end_date' and 'window_in_years' must be provided."
            )

        values.start_date = utilities.calculate_start_date(end_date=values.end_date, window_in_years=values.window_in_years)
        return values

class ComparableRequest(ComparableBaseRequest):
    peccs_country_profile: typing.Optional[PeccsCountryProfile] = None 
    factors_profiles: typing.Optional[FactorProfile] = None
    operation: typing.Optional[str] = None
    intersect_peccs: typing.Optional[bool] = None   
   
    def to_dict(self):
        payload = self.model_dump(mode="json", by_alias=True, exclude={"factors_profiles"}, exclude_defaults=True)
        if self.factors_profiles:
            payload["factorsProfiles"] = self.factors_profiles.to_list(currency=self.currency)

        return payload

class ComparableBoundaryRequest(ComparableBaseRequest):
    factor_name: str
    peccs_country_profile: typing.Optional[PeccsCountryProfile] = None 

    def to_dict(self):
        payload = self.model_dump(mode="json", by_alias=True, exclude_defaults=True)
        return payload


import pydantic
import typing
import sipametrics.models.base as base_models


class MetricsRequest(base_models.CamelCaseBaseModel):
    entity_id: typing.Union[list[str], str] = pydantic.Field(serialization_alias="indexIds")
    metric_id: typing.Union[list[str], str] = pydantic.Field(serialization_alias="typeIds")

    @pydantic.field_validator("entity_id", "metric_id", mode="before")
    @classmethod
    def ensure_list(cls, v):
        if isinstance(v, str):
            return [v]
        return v
    
    def to_dict(self) -> dict:
        payload = { "metrics": [ self.model_dump(by_alias=True) ] }
        return payload



import datetime
import dateutil.relativedelta
import typing


def calculate_start_date(end_date: typing.Optional[datetime.date], window_in_years: typing.Optional[int]) -> typing.Optional[datetime.date]:
    if not end_date or not window_in_years:
        return None

    start_date = end_date - dateutil.relativedelta.relativedelta(years=window_in_years)
    return start_date



import datetime
import pydantic
import typing
import sipametrics.internal.markers as _markers
import sipametrics.internal.endpoints as _endpoints
import sipametrics.internal.clients as _clients
import sipametrics.models.metrics as metrics_models
import sipametrics.models.pi_comparables as pi_comparables_models, sipametrics.models.pe_comparables as pe_comparables_models
import sipametrics.models.term_structures as term_structures_models
import sipametrics.models.indices_catalogue as indices_catalogue_models
import sipametrics.models.taxonomies as taxonomies_models
import sipametrics.models.custom_benchmarks as custom_benchmarks_models
import sipametrics.models.direct_alpha as direct_alpha_models
import sipametrics.constants as CONSTANTS


class SipaMetricsService:
    def __init__(self, api_key: str, api_secret: str):
        self.client = _clients.BaseClient(api_key, api_secret)

    async def close(self):
        await self.client.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, tb):
        await self.client.close()


    ######
    # Metrics
    ######
    async def metrics(self, 
                      entity_id: typing.Union[list[str], str], 
                      metric_id: typing.Union[list[str], str]) -> dict:
        """
        Returns data for metrics of a given index or a list of indices. Refer to https://docs.sipametrics.com/docs/2-2-1-quick-start for details.

        Args:
            entity_id (str or list of str): One or more index tickers.
            metric_id (str or list of str): One or more metric tickers.

        Returns:
            dict: Results wrapped in 'data' element.

        Raises:
            ValueError: If one or more arguments contain invalid or unsupported values.
        """
        try:
            request = metrics_models.MetricsRequest(entity_id=entity_id, metric_id=metric_id)
        except pydantic.ValidationError as e:
            raise ValueError(f"Invalid arguments: {e}")
        
        params = request.to_dict()
        return await self.client._post(_endpoints.URLS[self.metrics.__name__], data=params)    


    ######
    # PI Comparables
    ######
    async def infra_equity_comparable(
        self,
        metric: str,
        currency: typing.Optional[str] = None,
        age_in_months: typing.Optional[int] = None,
        end_date: typing.Optional[datetime.date] = None,
        window_in_years: typing.Optional[int] = None,
        industrial_activities: typing.Optional[list[str]] = None,
        business_risk: typing.Optional[str] = None,
        corporate_structure: typing.Optional[str] = None,
        countries: typing.Optional[list[str]] = None,
        size: typing.Optional[typing.Union[float, str]] = None,
        leverage: typing.Optional[typing.Union[float, str]] = None,
        profitability: typing.Optional[typing.Union[float, str]] = None,
        investment: typing.Optional[typing.Union[float, str]] = None,
        time_to_maturity: typing.Optional[typing.Union[float, str]] = None,
        factor_weight: typing.Optional[float] = None,
        intersect_ticcs: typing.Optional[bool] = None,
    ) -> dict:
        """
        Performs a comparable computation for private infrastructure equities. This involves finding data points which have similar TICCS classifications and factor values,
        and averaging the metric values. Refer to https://docs.sipametrics.com/docs/2-3-2-sipametrics-infra_equity_comparable for more details.

        Args:
            metric (str): The metric for which the comparable has to be evaluated.
            currency (str, optional): The currency of monetary factor inputs such as size.
            age_in_months (int, optional): The age of the company in months. If this parameter is set, end_date and window_in_years will be ignored.
            end_date (date, optional): The maximum date of the comparable dataset.
            window_in_years (int, optional): The window in years of the comparable dataset.
            industrial_activities (list of str, optional): List of industrial activity TICCS codes.
            business_risk (str, optional): Business risk TICCS code.
            corporate_structure (str, optional): Corporate structure TICCS code.
            countries (list of str, optional): List of country ISO codes.
            size (float or str, optional): Total assets or size of the company. Represented either as an absolute value in millions of the specified currency or as a
                quintile value.
            leverage (float or str, optional): Total senior liabilities over total assets. Represented either as a percentage or as a quintile value.
            profitability (float or str, optional): Return on assets or profitability metric. Represented either as a percentage or as a quintile value.
            investment (float or str, optional): Capital expenditures over total assets. Represented either as a percentage or as a quintile value.
            time_to_maturity (float or str, optional): Years until maturity. Represented as either a numerical value or as a bucket.
            factor_weight (float, optional):
            intersect_ticcs (bool, optional): 

        Returns:
            dict: Results wrapped in 'data' element.

        Raises:
            ValueError: If one or more arguments contain invalid or unsupported values.
        """
        try:
            ticcs_profile = pi_comparables_models.TiccsProfile(
                industrial_activities=industrial_activities,
                business_risk=business_risk,
                corporate_structure=corporate_structure
            )
            factors_profile = pi_comparables_models.EquityFactorProfile(
                currency=currency,
                countries=countries,
                size=size,
                leverage=leverage,
                profitability=profitability,
                investment=investment,
                time_to_maturity=time_to_maturity
            )
            request = pi_comparables_models.ComparableRequest(
                metric=metric,
                age_in_months=age_in_months,
                end_date=end_date,
                window_in_years=window_in_years,
                ticcs_profiles=ticcs_profile,
                factors_profiles=factors_profile,
                factor_weight=factor_weight,
                intersect_ticcs=intersect_ticcs,
            )
        except pydantic.ValidationError as e:
            raise ValueError(f"Invalid arguments: {e}")
        
        params = request.to_dict()
        return await self.client._post(_endpoints.URLS[self.infra_equity_comparable.__name__], data=params)    

    async def infra_debt_comparable(
        self,
        metric: str,
        currency: typing.Optional[str] = None,
        age_in_months: typing.Optional[int] = None,
        end_date: typing.Optional[datetime.date] = None,
        window_in_years: typing.Optional[int] = None,
        industrial_activities: typing.Optional[list[str]] = None,
        business_risk: typing.Optional[str] = None,
        corporate_structure: typing.Optional[str] = None,
        countries: typing.Optional[list[str]] = None,
        face_value: typing.Optional[typing.Union[float, str]] = None,
        time_to_maturity: typing.Optional[typing.Union[float, str]] = None,
        factor_weight: typing.Optional[float] = None,
        intersect_ticcs: typing.Optional[bool] = None,        
    ) -> dict:
        """
        Perform a comparable computation for private infrastructure debts. This involves finding datapoints which have similar TICCS classifications and factor values,
        and averaging the metric values. Refer to https://docs.sipametrics.com/docs/2-3-3-sipametrics-infra_debt_comparable for more information.

        Args:
            metric (str): The metric for which the comparable has to be evaluated.
            currency (str, optional): The currency of monetary factor inputs such as size.
            age_in_months (int, optional): The age of the company in months.
            end_date (date, optional): The maximum date of the comparable dataset.
            window_in_years (int, optional): The window in years of the comparable dataset.
            industrial_activities (list of str, optional): List of industrial activity TICCS codes.
            business_risk (str, optional): Business risk TICCS code.
            corporate_structure (str, optional): Corporate structure TICCS code.
            countries (list of str, optional): List of country ISO codes.
            face_value (float or str, optional): Face value of assets, represented either as an absolute value in millions of the specified currency or as a
                quintile value.
            time_to_maturity (float or str, optional): Years until maturity. Represented as either a numerical value or as a bucket.
            factor_weight (float, optional):
            intersect_ticcs (bool, optional): 

        Returns:
            dict: Results wrapped in 'data' element.

        Raises:
            ValueError: If one or more arguments contain invalid or unsupported values.
        """
        try:
            ticcs_profile = pi_comparables_models.TiccsProfile(
                industrial_activities=industrial_activities,
                business_risk=business_risk,
                corporate_structure=corporate_structure
            )
            factors_profile = pi_comparables_models.DebtFactorProfile(
                currency=currency,
                countries=countries,
                face_value=face_value,
                time_to_maturity=time_to_maturity
            )  
            request = pi_comparables_models.ComparableRequest(
                metric=metric,
                age_in_months=age_in_months,
                end_date=end_date,
                window_in_years=window_in_years,
                ticcs_profiles=ticcs_profile,
                factors_profiles=factors_profile,
                factor_weight=factor_weight,
                intersect_ticcs=intersect_ticcs,
            )      
        except pydantic.ValidationError as e:
            raise ValueError(f"Invalid arguments: {e}")
        
        params = request.to_dict()
        return await self.client._post(_endpoints.URLS[self.infra_debt_comparable.__name__], data=params)  


    ######
    # Term Structure
    ######
    async def term_structure(self, country: str, date: datetime.date, maturity_date: datetime.date) -> dict:
        """
        Query annualised risk-free rate for a given country and maturity date on the curve. Refer to https://docs.sipametrics.com/docs/2-3-4-sipametrics-termstructure for
        more information.

        Args:
            country (str): Three-letter ISO code representing the country.
            date (date): Date in point.
            maturity_date (date): Maturity date of the term structure.

        Returns:
            dict: Results wrapped in 'data' element.

        Raises:
            ValueError: If one or more arguments contain invalid or unsupported values.
        """
        try:
            request = term_structures_models.TermStructureRequest(
                country=country,
                date=date,
                maturity_date=maturity_date
            )
        except pydantic.ValidationError as e:
            raise ValueError(f"Invalid arguments: {e}")
        
        params = request.to_dict()
        return await self.client._post(_endpoints.URLS[self.term_structure.__name__], data=params)


    ######
    # PE Comparables
    ######
    async def private_equity_comparable(
        self,
        metric: str,
        currency: typing.Optional[str] = None,
        age_in_months: typing.Optional[int] = None,
        end_date: typing.Optional[datetime.date] = None,
        window_in_years: typing.Optional[int] = None,
        industrial_activities: typing.Optional[list[str]] = None,
        revenue_models: typing.Optional[list[str]] = None,
        customer_models: typing.Optional[list[str]] = None,
        lifecycle_phases: typing.Optional[list[str]] = None,
        value_chain_types: typing.Optional[list[str]] = None,
        countries: typing.Optional[list[str]] = None,
        size: typing.Optional[typing.Union[float, str]] = None,
        growth: typing.Optional[typing.Union[float, str]] = None,
        leverage: typing.Optional[typing.Union[float, str]] = None,
        profits: typing.Optional[typing.Union[float, str]] = None,
        country_risk: typing.Optional[list[str]] = None,
        universe: typing.Optional[str] = None,
        factor_weight: typing.Optional[float] = None,
        type: typing.Optional[str] = "mean",
        intersect_peccs: typing.Optional[bool] = None,
    ) -> dict:
        """
        Perform a comparable computation for private equities. This involves finding datapoints which have similar PECCS classifications and factor values and
        averaging the metric values. Refer to https://docs.sipametrics.com/docs/2-3-5-sipametrics-private_equity_comparable for more information.

        Args:
            metric (str): The metric for which the comparable has to be evaluated.
            currency (str, optional): The currency of monetary factor inputs such as size.
            age_in_months (int, optional): The age of the company in months.
            end_date (date, optional): The maximum date of the comparable dataset.
            window_in_years (int, optional): The window in years of the comparable dataset. 
            industrial_activities (list of str, optional): List of industrial activity PECCS codes.
            revenue_models (list of str, optional): List of revenue model PECCS codes.
            customer_models (list of str, optional): List of customer model PECCS codes.
            lifecycle_phases (list of str, optional): List of lifecycle phase PECCS codes.
            value_chain_types (list of str, optional): List of value chain type PECCS codes.
            countries (list of str, optional): List of country ISO codes.
            size (float or str, optional): Revenue of the company. Represented either as an absolute value in millions of the specified currency or as a quintile value.
            growth (float or str, optional): Revenue growth. Represented either as a percentage or as a quintile value.
            leverage (float or str, optional): Total senior liabilities over total assets. Represented either as a percentage or as a quintile value.
            profits (float or str, optional): EBITDA margin. Represented either as a percentage or as a quintile value.
            country_risk (list of str, optional): Term spread. Specified as a list of country ISO codes.
            universe (str, optional): Universe of companies in the dataset.
            factor_weight (float, optional): A decimal value between 0 and 1. At the extremes, 1 indicates that comparables are purely based on factors, while 0
                indicates that comparables are purely based on PECCS. Values between 0 and 1 create a weighted average between the two.
            type (str, optional): Determines how to aggregate the comparables dataset. Default is "mean".
            intersect_peccs (bool, optional): Determines whether to intersect the PECCS codes. Default is True.

        Returns:
            dict: Results wrapped in 'data' element.

        Raises:
            ValueError: If one or more arguments contain invalid or unsupported values.
        """
        try:
            peccs_country_profile = pe_comparables_models.PeccsCountryProfile(
                industrial_activities=industrial_activities,
                revenue_models=revenue_models,
                customer_models=customer_models,
                lifecycle_phases=lifecycle_phases,
                value_chain_types=value_chain_types,
                countries=countries
            )
            factors_profile = pe_comparables_models.FactorProfile(
                size=size,
                growth=growth,
                leverage=leverage,
                profits=profits,
                country_risk=country_risk
            )
            request = pe_comparables_models.ComparableRequest(
                metric=metric,
                age_in_months=age_in_months,
                end_date=end_date,
                window_in_years=window_in_years,
                peccs_country_profile=peccs_country_profile,
                factors_profiles=factors_profile,
                currency=currency,
                universe=universe,
                factor_weight=factor_weight,
                operation=type,
                intersect_peccs=intersect_peccs
            )
        except pydantic.ValidationError as e:
            raise ValueError(f"Invalid arguments: {e}")
        
        params = request.to_dict()
        return await self.client._post(_endpoints.URLS[self.private_equity_comparable.__name__], data=params)
 
    async def private_equity_comparable_boundaries(
        self,
        metric: str,
        factor_name: str,
        currency: typing.Optional[str] = None,
        age_in_months: typing.Optional[int] = None,
        end_date: typing.Optional[datetime.date] = None,
        window_in_years: typing.Optional[int] = None,
        industrial_activities: typing.Optional[list[str]] = None,
        revenue_models: typing.Optional[list[str]] = None,
        customer_models: typing.Optional[list[str]] = None,
        lifecycle_phases: typing.Optional[list[str]] = None,
        value_chain_types: typing.Optional[list[str]] = None,
        countries: typing.Optional[list[str]] = None,
        universe: typing.Optional[str] = None
    ) -> dict:
        """
        Evaluates the quintile boundaries for a given metric within the private equity comparables dataset. Refer to 
        https://docs.sipametrics.com/docs/2-3-6-sipametrics-private_equity_comparable_bounda for more information.

        Args:
            metric (str): The metric for which the comparable has to be evaluated.
            factor_name (str): The factor for which to obtain the quintile boundaries.
            currency (str, optional): The currency of monetary factor inputs, such as size.
            age_in_months (int, optional): The age of the company in months. 
            end_date (date, optional): The maximum date of the comparable dataset.
            window_in_years (int, optional): The window in years of the comparable dataset.
            industrial_activities (list of str, optional): List of industrial activity PECCS codes.
            revenue_models (list of str, optional): List of revenue model PECCS codes.
            customer_models (list of str, optional): List of customer model PECCS codes.
            lifecycle_phases (list of str, optional): List of lifecycle phase PECCS codes.
            value_chain_types (list of str, optional): List of value chain type PECCS codes.
            countries (list of str, optional): List of country ISO codes.
            universe (str, optional): Universe of companies in the dataset.
        
        Returns:
            dict: Results wrapped in 'data' element.

        Raises:
            ValueError: If one or more arguments contain invalid or unsupported values.
        """
        try:
            peccs_country_profile = pe_comparables_models.PeccsCountryProfile(
                industrial_activities=industrial_activities,
                revenue_models=revenue_models,
                customer_models=customer_models,
                lifecycle_phases=lifecycle_phases,
                value_chain_types=value_chain_types,
                countries=countries
            )
            request = pe_comparables_models.ComparableBoundaryRequest(
                metric=metric,
                currency=currency,
                factor_name=factor_name,
                age_in_months=age_in_months,
                end_date=end_date,
                window_in_years=window_in_years,
                peccs_country_profile=peccs_country_profile,
                universe=universe
            )
        except pydantic.ValidationError as e:
            raise ValueError(f"Invalid arguments: {e}")
        
        params = request.to_dict()
        return await self.client._post(_endpoints.URLS[self.private_equity_comparable_boundaries.__name__], data=params)


    ######
    # Incides Catalogue
    ######
    async def indices_catalogue(
        self, 
        product: CONSTANTS.Product, 
        app: typing.Optional[CONSTANTS.App] = None
    ) -> dict:
        """
        List of indices, as well as the corresponding categories and tickers. Refer to https://docs.sipametrics.com/docs/2-3-7-sipametrics-indices_catalogue for
        more information.

        Args:
            product (str): Type of product.
            app (App, optional): App within the product. 

        Returns:
            dict: Results wrapped in 'data' element.

        Raises:
            ValueError: If one or more arguments contain invalid or unsupported values.
        """
        try:
            request = indices_catalogue_models.IndicesCatalogueRequest(product=product, app=app)
        except pydantic.ValidationError as e:
            raise ValueError(f"Invalid arguments: {e}")
        
        params = request.to_dict()
        return await self.client._get(_endpoints.URLS[self.indices_catalogue.__name__], params=params)
    

    ######
    # Taxonomies
    ######
    async def taxonomies(
            self, 
            taxonomy: CONSTANTS.Taxonomy, 
            pillar: typing.Optional[typing.Union[CONSTANTS.Ticcs, CONSTANTS.Peccs]] = None
    ) -> dict:
        """
        Retrieves a list of supported taxonomy (TICCS®, PECCS®, and TICCS®+) definitions and the relevant mappings. Refer to 
        https://docs.sipametrics.com/docs/2-3-8-sipametrics-taxonomies for more information.

        Args:
            taxonomy (str): The taxonomy to query.
            pillar (Ticcs, Peccs, optional): The pillar within the taxonomy to be queried.

        Returns:
            dict: Results wrapped in 'data' element.

        Raises:
            ValueError: If one or more arguments contain invalid or unsupported values.
        """
        try:
            request = taxonomies_models.TaxonomiesRequest(taxonomy=taxonomy, pillar=pillar)
        except pydantic.ValidationError as e:
            raise ValueError(f"Invalid arguments: {e}")

        params = request.to_dict()
        destined_url = _endpoints.URLS[self.taxonomies.__name__].format(taxonomy=params["taxonomy"])
        return await self.client._get(destined_url, params={ "pillar": params["pillar"] })


    ######
    # PE Custom Benchmarks
    ######
    @_markers.beta
    async def private_equity_custom_benchmarks(
        self,
        index_ticker: str,
        start_date: typing.Optional[datetime.date] = None,
        end_date: typing.Optional[datetime.date] = None,
        currency: typing.Optional[str] = None,
        constraints: typing.Optional[list[typing.Union[custom_benchmarks_models.Constraint, custom_benchmarks_models.EqualityConstraint, custom_benchmarks_models.RangeConstraint]]] = None,
        initial_index_value: typing.Optional[float] = None,
        allocations: typing.Optional[bool] = None,
        risk_profiles: typing.Optional[bool] = None,
        include_constituents: typing.Optional[bool] = None,
    ) -> dict:
        """
        To generate custom benchmarks based on an existing private equity index, with the option to modify the weightage of constituents by constraints. Refer to 
        https://docs.sipametrics.com/docs/2-3-9-private_equity_custom_benchmarks for more information.

        Args:
            index_ticker (str): Ticker of the index.
            start_date (date, optional): Start date of the index.
            end_date (date, optional): End date of the index.
            currency (str, optional): The reporting currency.
            constraints (list of EqualityConstraint, list of RangeConstraint, list of Constraint, optional): List of constraints to be applied to the index.
            initial_index_value (float, optional): Initial index value.
            allocations (bool, optional): Whether to include PECCS allocations. Default is False.
            risk_profiles (bool, optional): Whether to include risk profiles. Default is False.
            include_constituents (bool, optional): Whether to include list of constituents. Default is False.

        Returns:
            dict: Results wrapped in 'data' element.

        Raises:
            ValueError: If one or more arguments contain invalid or unsupported values.
        """
        try:
           request = custom_benchmarks_models.CustomBenchmarkRequest(
                external_ticker=index_ticker,
                start_date=start_date,
                end_date=end_date,
                currency=currency,
                constraints=constraints,
                initial_index_value=initial_index_value,
                allocations=allocations,
                risk_profiles=risk_profiles,
                include_constituents=include_constituents
            )
        except pydantic.ValidationError as e:
            raise ValueError(f"Invalid arguments: {e}")

        params = request.to_dict()
        return await self.client._post(_endpoints.URLS[self.private_equity_custom_benchmarks.__name__], data=params) 


    ######
    # PE Region Tree
    ######
    async def private_equity_region_tree(self) -> dict:
        """
        Returns hierarchical view of regions, subregions and countries within private equity universe. Refer to
        https://docs.sipametrics.com/docs/2-3-10-sipametrics-private_equity_region_tree for more information.

        Returns:
            dict: Results wrapped in 'data' element.
        """
        return await self.client._get(_endpoints.URLS[self.private_equity_region_tree.__name__])


    ######
    # PI Custom Benchmarks
    ######
    @_markers.beta
    async def private_infra_custom_benchmarks(
        self,
        index_ticker: str,
        start_date: typing.Optional[datetime.date] = None,
        end_date: typing.Optional[datetime.date] = None,
        currency: typing.Optional[str] = None,
        constraints: typing.Optional[list[typing.Union[custom_benchmarks_models.Constraint, custom_benchmarks_models.EqualityConstraint, custom_benchmarks_models.RangeConstraint]]] = None,
        initial_index_value: typing.Optional[float] = None,
        allocations: typing.Optional[bool] = None,
        risk_profiles: typing.Optional[bool] = None,
        include_constituents: typing.Optional[bool] = None,
    ) -> dict:
        """
        To generate custom benchmarks based on an existing private infrastructure index, with the option to modify the weightage of constituents by constraints. 

        Args:
            index_ticker (str): Ticker of the index.
            start_date (date, optional): Start date of the index.
            end_date (date, optional): End date of the index.
            currency (str, optional): The reporting currency.
            constraints (list of Constraint, optional): List of constraints to be applied to the index.
            initial_index_value (float, optional): Initial index value.
            allocations (bool, optional): Whether to include TICCS allocations.
            risk_profiles (bool, optional): Whether to include risk profiles.
            include_constituents (bool, optional): Whether to include list of constituents.

        Returns:
            dict: Results wrapped in 'data' element.

        Raises:
            ValueError: If one or more arguments contain invalid or unsupported values.
        """
        try:
           request = custom_benchmarks_models.CustomBenchmarkRequest(
                external_ticker=index_ticker,
                start_date=start_date,
                end_date=end_date,
                currency=currency,
                constraints=constraints,
                initial_index_value=initial_index_value,
                allocations=allocations,
                risk_profiles=risk_profiles,
                include_constituents=include_constituents
            )
        except pydantic.ValidationError as e:
            raise ValueError(f"Invalid arguments: {e}")

        params = request.to_dict()
        return await self.client._post(_endpoints.URLS[self.private_infra_custom_benchmarks.__name__], data=params) 


    ######
    # Direct Alpha
    ######    
    @_markers.beta
    async def direct_alpha(
        self,
        data: list[direct_alpha_models.FundAlphaData]
    ) -> dict:
        """
        Produce alpha comparison for a fund against a broadmarket index and optionally a benchmark index.

        Args:
            data (list of FundAlphaData): External ticker of the index.

        Returns:
            dict: Results wrapped in 'data' element.

        Raises:
            ValueError: If one or more arguments contain invalid or unsupported values.
        """
        try:
            request = direct_alpha_models.AlphaRequest(
                fund_alpha_data=data,
            )
        except pydantic.ValidationError as e:
            raise ValueError(f"Invalid arguments: {e}")

        params = request.to_dict()
        return await self.client._post(_endpoints.URLS[self.direct_alpha.__name__], data=params) 



import pydantic
import typing
import sipametrics.models.base as base_models
import sipametrics.constants as CONSTANTS


class TaxonomiesRequest(base_models.CamelCaseBaseModel):
    taxonomy: CONSTANTS.Taxonomy
    pillar: typing.Optional[typing.Union[CONSTANTS.Ticcs, CONSTANTS.Peccs]] = None

    model_config = pydantic.ConfigDict(use_enum_values=True)


    @pydantic.model_validator(mode="after")
    def validate_pillar_in_taxonomy(cls, values):
        if values.pillar:
            if values.taxonomy in [CONSTANTS.Taxonomy.TICCS_TO_NACE, CONSTANTS.Taxonomy.TICCS_TO_EU]:
                raise ValueError("Pillar is not supported for TICCS-to-NACE or TICCS-to-EU taxonomy.")
            if values.taxonomy == CONSTANTS.Taxonomy.TICCS and values.pillar not in {pillar for pillar in CONSTANTS.Ticcs}:
                raise ValueError("For TICCS taxonomy, pillar has to be one of the values in CONSTANTS.Ticcs.")
            if values.taxonomy == CONSTANTS.Taxonomy.PECCS and values.pillar not in {pillar for pillar in CONSTANTS.Peccs}:
                raise ValueError("For PECCS taxonomy, pillar has to be one of the values in CONSTANTS.Peccs.")
            if values.taxonomy == CONSTANTS.Taxonomy.TICCS_PLUS and values.pillar not in {pillar for pillar in CONSTANTS.TiccsPlus}:
                raise ValueError("For TICCS+ taxonomy, pillar has to be one of the values in CONSTANTS.TiccsPlus.")
            
        return values

    def to_dict(self) -> dict:
        return self.model_dump(by_alias=True)


import functools
import warnings


def beta(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        warnings.warn(
            f"Function '{func.__name__}' is in BETA and may change in future versions.",
            category=UserWarning,
            stacklevel=2
        )
        return func(*args, **kwargs)
    return wrapper
